/* 
 * easter.c   Calculates easter sunday          1999-2-13
 * Copyright (C) 2000 Antonio Luque Estepa  <aluque@zipi.us.es>
 * Copyright (C) 2001 Amaury Jacquot <sxpert@esitcom.org>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#if(!defined HAVE_GETOPT_H)
#error Must have <getopt.h>
#endif

#include <stdio.h>
#include <getopt.h>
#include <ctype.h>
#include <sys/time.h>
#include <time.h>
#include <unistd.h>
#include <libintl.h>

#define _(String) gettext(String)
#define gettext_noop(String) (String)
#define N_(String) gettext_noop(String)

/* For debugging purposes only */
#define DEBUG

#define SUNDAY 0
#define MONDAY 1
#define TUESDAY 2
#define WEDNESDAY 3
#define THURSDAY 4
#define FRIDAY 5
#define SATURDAY 6

#define isleap(yr) \
(((yr) <= 1582 ) ? !((yr)%4) : \
(!((yr)%4)) && (((yr)%100) || !((yr)%400)))

/* Number of julian leap years that are not gregorian leap years since 1582 */
#define notleapsince1582(yr) \
(((yr) >= 1600 ? (yr)/100 -16: 0) -((yr) >= 2000 ? (yr)/400 -4: 0)) 

/* Number of leap years since year 1 */
#define leapsince1(yr) \
( (yr)/4 - notleapsince1582(yr) )

#define JULIAN5OCT1582 577738
#define JD0 2415020L

int main(int, char **);
void usage(void);
void version(void);
int dayofweek(int,int,int);
int jday(int,int,int);
void juliantodate(int,int,int *,int *);
int yeartoeaster(int *,int *,int);
int daysto(int,int,int);
double juliannr(int,int,int);
void printdate(FILE*,int,int,int);
/* this function allocates a string that must be freed by caller */
char* datetostring(int,int,int);
char* dprint(char*,int,int,int);

char const progname[]="easter 2.2.2  (c) 2000 Antonio Luque <aluque@zipi.us.es>\n"
                      "              (c) 2001 Amaury Jacquot <sxpert@esitcom.org>\n";
char usagemesg[]=N_("easter [-rac] [-f] [-hvVi] year\n"
"easter calculates easter of given year.\n"
"\t-c Calculates Ash Wednesday.\n"
"\t-r Calculates Palm Sunday.\n"
"\t-e Calculates Easter Sunday. (same as nothing)\n"
"\t-l Calculates Easter monday.\n"
"\t-s Calculates Ascension.\n"
"\t-p Calculates Pentecost.\n"
"\t-m Calculates Pentecost monday.\n"
"\n"
"\t-a Displays all dates for script uses.\n"
"\t-A Displays all dates and their names.\n"
"\t-f Number of days remaining to Easter (or Palm, or Ash).\n"
"\t-V Verbose mode.\n"
"\t-W Extra-verbose mode.\n"
"\t-v Shows version number and exits\n"
"\t-h Shows this text and exits\n"
"If year is not given, use current.\n"
"if multiple dates are entered, they are printed in the order above.\n"
"If no option, dates displayed in day month format.\n"
"If -i option, displayed in year-month-day (ISO 8601) format.\n");

int verbose;
/* Note that ISO dates are of the form yyyy-mm-dd and traditional dates
 * are of the form dd mm (no hyphens).                                   */
int isodate;
int remain;

typedef struct {
  int day;
  int month;
  int year;
} date;

/* Table of Paschal full moons depending on the epacta */
date paschal[]={
  {0,0}, {12,4}, {11,4}, {10,4}, {9,4}, {8,4}, {7,4}, {6,4}, {5,4}, {4,4},
  {3,4}, {2,4},  {1,4},  {31,3}, {30,3},{29,3},{28,3},{27,3},{26,3},{25,3},
  {24,3},{23,3}, {22,3}, {21,3}, {18,4},{0,0}, {17,4},{16,4},{15,4},{14,4},
  {13,4}
};
/* If epacta==25, paschal moon is determined by golden number */

/* Number of days in each month */
int monthday[]=
{ 31,28,31,30,31,30,31,31,30,31,30,31 };
int leapmonthday[]=
{ 31,29,31,30,31,30,31,31,30,31,30,31 };

char *dayname[7]= { N_("Sunday"),
		    N_("Monday"),
		    N_("Tuesday"),
		    N_("Wednesday"),
		    N_("Thursday"),
		    N_("Friday"),
		    N_("Saturday") };

#define MAX_HOLIDAYS 12

/* dates should be in chronological order (makes it easier) */
char *holiday_name[MAX_HOLIDAYS]= { N_("Ash Wednesday"),
				    N_("I   Lent"),
				    N_("II  Lent"),
				    N_("III Lent"),
				    N_("IV  Lent"),
				    N_("V   Lent"),
				    N_("Palm Sunday"),
				    N_("Easter Sunday"),
				    N_("Easter Monday"),
				    N_("Ascension"),
				    N_("Pentecost Sunday"),
				    N_("Pentecost Monday") };
enum {
  ASH_WEDNESDAY,
  LENT_1,
  LENT_2,
  LENT_3,
  LENT_4,
  LENT_5,
  PALM_SUNDAY,
  EASTER_SUNDAY,
  EASTER_MONDAY,
  ASCENSION,
  PENTECOST_SUNDAY,
  PENTECOST_MONDAY,
};

#define ALL 0xFFF

int main(int argc, char *argv[])
{
  int c,i,pos,bit,date_remaining;
  char options[]="c12345relspmhvVWaAfi";
  int year;
  date dates[MAX_HOLIDAYS]; /* stores all dates */
  date this_date;
  int easter_jd;  /* Easter Julian day */
  int calc_map=0;
  int all=0;
  int labels=0;
  struct timeval tv;
  struct tm *ttt;
  int sum_bits;
  char* str;
  
  setlocale(LC_ALL,"");
  bindtextdomain(PACKAGE,LOCALEDIR);
  textdomain(PACKAGE);
  
  if(argc<1) { /* Wrong number of arguments */
    usage();
    exit(0);
  }
  
  /* Parse command line options */
  while((c=getopt(argc,argv,options))!=-1){
    /* handles all holidays */
    pos=((long)strchr(options,c)-(long)options);
    if (pos<MAX_HOLIDAYS) {
      calc_map|=1<<pos;
    } else {
      switch(c) {
      case 'A':
	all=1;
      case 'a':
	calc_map|=ALL;
	break;
      case 'f':
	remain=1;
	break;	    
      case 'i':
	isodate=1;
	break;
      case 'W':
	verbose=1;
      case 'V':
	labels=1;
	break;
      case 'v':
	version();
	exit(1);
      case 'h':
      case '?':
	usage();
	exit(1);
      default:
	abort();
      }
    }
  }
  if (calc_map==0) calc_map|=1<<EASTER_SUNDAY;

  if(optind==argc) { /* No year was specified */
    gettimeofday(&tv,NULL);
    ttt=localtime((time_t *)&tv.tv_sec);
    year=ttt->tm_year+1900;
  }
  else {
    if(argv[optind][0]=='-') {
      if(!isdigit(argv[optind][1])) {
	printf(_("easter: Can not understand year.\n"));
	exit(0);
      }
    }
    else if(!isdigit(argv[optind][0])) {
      printf(_("easter: Can not understand year.\n"));
      exit(0);	    
    }
    sscanf(argv[optind],"%d",&year);
  }
  if(year==0) {
    fprintf(stderr,_("easter: there was no year 0.\n"));
    exit(0);
  }
  
  /* Calculate all relevant dates */
  easter_jd=yeartoeaster(&dates[EASTER_SUNDAY].day,&dates[EASTER_SUNDAY].month,year);
  juliantodate(easter_jd+50,year,&dates[PENTECOST_MONDAY].day,&dates[PENTECOST_MONDAY].month);
  juliantodate(easter_jd+49,year,&dates[PENTECOST_SUNDAY].day,&dates[PENTECOST_SUNDAY].month);
  juliantodate(easter_jd+39,year,&dates[ASCENSION].day,&dates[ASCENSION].month);
  juliantodate(easter_jd+1,year,&dates[EASTER_MONDAY].day,&dates[EASTER_MONDAY].month);
  juliantodate(easter_jd-7,year,&dates[PALM_SUNDAY].day,&dates[PALM_SUNDAY].month);
  juliantodate(easter_jd-46,year,&dates[ASH_WEDNESDAY].day,&dates[ASH_WEDNESDAY].month);
  for(i=0;i<5;++i)
    juliantodate(easter_jd-7*(i+2),year,&dates[LENT_5-i].day,&dates[LENT_5-i].month);
  
  if(all)  {
    printf(_("Calendar for %04d:\n"),year);
    for(i=0;i<MAX_HOLIDAYS;i++) {
      str=datetostring(year,dates[i].month,dates[i].day);
      printf(_("  %s\t\t%s\n"),_(holiday_name[i]),str);
      free(str);
    }
  } else {
    for(i=0;i<MAX_HOLIDAYS;i++) {
      if ((calc_map>>i)&1) {
	str=datetostring(year,dates[i].month,dates[i].day);
	if(labels) printf(_("%s for year %d: %s\n"),_(holiday_name[i]),year,str);
	else printf("%s\n",str);
	free(str);
      }
    }
  } 
}

void usage(void)
{
  fprintf(stdout,progname);
  fprintf(stdout,_(usagemesg));
}

void version(void)
{
  fprintf(stdout,progname);
}

/* dayofweek returns the day of the week, given the complete date */
int dayofweek(int day,int month,int year)
{
  long julianday; /* Julian day total since 1 Jan 1 */
  int weekday=THURSDAY;
  char* str;
  
  julianday=(year-1)*365+leapsince1(year-1)+jday(day,month,year);
  
  if(julianday < JULIAN5OCT1582) { /* Julian calendar */
    weekday=((julianday-1 + SATURDAY)%7);
  }
  
  if(julianday >= JULIAN5OCT1582+9){
    weekday=((julianday-1+SATURDAY - 10)%7);
  }
  
  if(verbose) {
    str=datetostring(year,month,day);
    fprintf(stderr,_("%s is %s\n"),str,_(dayname[weekday]));
    free(str);
  }
  
  return weekday;
}

/* Calculates the julian day (day number within the year) */
int jday(int day, int month,int year)
{
  int temp=0,i;
  if(isleap(year)){
    for(i=0;i<month-1;++i)
      temp+=leapmonthday[i];
    temp+=day;
  }
  else {
    for(i=0;i<month-1;++i)
      temp+=monthday[i];
    temp+=day;
  }
  return temp;
}

/* Converts a julian date to a normal date */
void juliantodate(int julian,int year, int *day,int *month)
{
  int i;
  if(isleap(year)) {
    for(i=0;i<12;++i){
      julian-=leapmonthday[i];
      if(julian<=0) {
	*month=i+1; *day=julian+leapmonthday[i];
	return;
      }
    }
  }
  else {   
    for(i=0;i<12;++i) {
      julian-=monthday[i];
      if(julian<=0) { 
	*month=i+1; *day=julian+monthday[i]; 
	return;
      }
    }
  }
}

/* Calculates easter for a given year
 *    day, month: Easter date
 *    year: year
 *  returns: julian day of easter in that year
 */
int yeartoeaster(int *day,int *month,int year)
{
  int cent;  /* Century */
  date full; /* Paschal Full moon */
  char* str;
  date easter;
  int gn,ep;  /* Golden Number, Epacta */
  int julianday,weekday;
  
  /* When will 21st century start? 
   * I dont want to dispute about the same again
   * For now, we assume century=two first digits of year + 1
   */
  cent=((year-(year%100))/100)+1;
  if(verbose) {
    fprintf(stderr,_("Calculating Easter for year %d\n"),year);
    if (isleap(year)) fprintf(stderr,_("Year is %d, and is leap. Century is %d\n"),year,cent);
    else fprintf(stderr,_("Year is %d, and is not leap. Century is %d\n"),year,cent);
  }
  
  gn=(year%19)+1;
  ep=(11*(gn-1))%30;
  ep=(int)(ep-(3*cent)/4);
  ep=(int)(ep+(8*cent+5)/25);
  ep+=8;
  while(ep<=0) ep+=30;
  while(ep>30) ep-=30;
  
  if(verbose) {
    fprintf(stderr,_("Golden Number is %d. Epacta is %d\n"),gn,ep);
  }
  
  full=paschal[ep];
  if(ep==25) {
    if(gn>11) { full.day=17; full.month=4; };
    if(gn<=11) { full.day=18; full.month=4; };
  }
  
  julianday=jday(full.day,full.month,year);   
  if(verbose) {
    str=datetostring(year,full.month,full.day);
    fprintf(stderr,_("Paschal full moon is %s, which is the %dth day of year\n"),str,julianday);
    free(str);
  }
  
  /* Now we must calculate the sunday following the paschal full moon */
  
  weekday=dayofweek(full.day,full.month,year);
  
  julianday+=7-weekday;
  juliantodate(julianday,year,&easter.day,&easter.month); 
  
  if(verbose) {
    str=datetostring(year,easter.month,easter.day);
    fprintf(stderr,_("Easter is %dth day of year, that is %s\n"),julianday,str);
    free(str);
  }
  
  *day=easter.day;
  *month=easter.month;
  return julianday;
}

/* Days remaining to a date
 *   day,month,year: date
 * returns: days remaining
 */
int daysto(int day,int month,int year)
{
  double jdtoday,jddate;
  date today;
  
  struct timeval tv;
  struct tm *ttt;
  
  gettimeofday(&tv,NULL);
  ttt=localtime((time_t *)&tv.tv_sec);
  today.year=ttt->tm_year+1900;
  today.day=ttt->tm_mday;
  today.month=ttt->tm_mon+1;
  jdtoday=juliannr(today.day,today.month,today.year);
  
  jddate=juliannr(day,month,year);
  
  return (int)(jddate-jdtoday);
}

/* Calculates Julian number for given date
 * This Julian number has year in consideration (is therefore, the official
 * astronomical Julian number). Note the difference with jday()
 *    day,month,year: date
 *  returns; Julian number
 * This function uses code based on:
 *  XEphem (c) 1990-97 Elwood Charles Downey <ecdowney@noao.edu>
 */
double juliannr(int day,int month,int year)
{
  int b,d,m,y;
  double JD;
  long c;
  char* str;
  
  m=month;
  y=(year < 0) ? year+1 : year;
  if(month<3) {
    m+=12;
    y-=1;
  }
  if(year<1582 || (year==1582 && (month<10  || (month==10 && day<15))))
    b=0;
  else {
    int a;
    a=y/100;
    b=2-a+a/4;
  }
  if(y<0)
    c=(long)((365.25*y) - 0.75) - 694025L;
  else
    c=(long)(365.25*y) - 694025L;
  
  d=30.6001*(m+1);
  JD=b+c+d+day;
  
  JD+=JD0;
  
  if(verbose) {
    str=datetostring(year,month,day);
    fprintf(stderr,_("Julian number for %s is %.0f\n"),str,JD);
    free(str);
  }
  return JD;
}

/* prints the date according to the isodate variable */
void printdate(FILE* f,int year,int month,int day) {
  char* str;
  str=datetostring(year,month,day);
  fprintf(f,"%s",str);
  free(str);
}

char* datetostring(int year,int month,int day) {
  char* str;
  unsigned int rem;
  if (isodate) {
    str=(char*)malloc(11);
    sprintf(str,"%04d-%02d-%02d",year,month,day);
  } else if(remain) {
    rem=daysto(day,month,year);
    str=(char*)malloc(12);
    sprintf(str,"%11d",rem);
  } else
    str=dprint(_("M-D"),year,month,day);
  return str;
}

char* dprint(char* format,int year,int month,int day) {
  int i,l=0;
  char* str;
  char tmp[5];
  
  if (format==NULL) return NULL;
  /* calculates the length of the string */
  for(i=0;i<strlen(format);i++) {
    switch(format[i]){
    case 'Y': l+=4; break;
    case 'M': l+=2; break;
    case 'm': 
      if (month<10) l++;
      else l+=2;
      break;
    case 'D': l+=2; break;
    case 'd': 
      if (day<10) l++;
      else l+=2;
      break;
    default: l++;
    }
  }
  str=(char*)malloc(l+1);
     
  /* prints the string */
  str[0]='\0';
  for(i=0;i<strlen(format);i++) {
    switch(format[i]){
    case 'Y': sprintf(tmp,"%04d",year); break;
    case 'M': sprintf(tmp,"%02d",month); break;
    case 'm': sprintf(tmp,"%d",month); break;
    case 'D': sprintf(tmp,"%02d",day); break;
    case 'd': sprintf(tmp,"%d",day); break;
    default: sprintf(tmp,"%c",format[i]);
    }
    strcat(str,tmp);
  }
  return str;
}
